from django.contrib.auth import get_user_model
from rest_framework import serializers


class ProfileSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = ['first_name', 'last_name']


def profile_serializer_handler():
    return ProfileSerializer



from ...views import LoginView, RegisterView, VerifyEmailView, EmailVerificationCompleteView, VerifyPhoneView, \
    PhoneVerificationCompleteView, ResendPhoneConfirmationView, ResendEmailConfirmationLinkView
from django.contrib.auth import views as auth_views
from django.urls import path, reverse_lazy

urlpatterns = [
    path('login/', LoginView.as_view(), name="login"),
    path('register/', RegisterView.as_view(), name='register'),
    path('logout/', auth_views.LogoutView.as_view(), name='logout'),

    # email urls
    path('verify/email/', ResendEmailConfirmationLinkView.as_view(), name='resend-email-confirmation'),
    path('verify/email/<str:uidb64>/<str:token>/', VerifyEmailView.as_view(), name='verify-email'),
    path('verify/email/complete', EmailVerificationCompleteView.as_view(), name='email-verification-complete'),

    # phone urls
    path('verify/phone/', VerifyPhoneView.as_view(), name='verify-phone'),
    path('verify/phone/complete', PhoneVerificationCompleteView.as_view(), name='phone-verification-complete'),
    path('resend_phone_activation/', ResendPhoneConfirmationView.as_view(), name='resend_phone_activation'),

    # password urls
    path('password_change/', auth_views.PasswordChangeView.as_view(), name='password_change'),
    path('password_change/done/', auth_views.PasswordChangeDoneView.as_view(), name='password_change_done'),

    path('password_reset/', auth_views.PasswordResetView.as_view(
        success_url=reverse_lazy('password_reset_done'),
        template_name='dj_accounts/password_reset_form.html',
        email_template_name='dj_accounts/password_reset_email.html',
        subject_template_name='dj_accounts/password_reset_subject.txt'
    ), name='password_reset'),

    path('password_reset/done', auth_views.PasswordResetDoneView.as_view(
        template_name='dj_accounts/password_reset_done.html',
    ), name='password_reset_done'),

    path('reset/<uidb64>/<token>', auth_views.PasswordResetConfirmView.as_view(
        template_name='dj_accounts/password_reset_confirm.html',
        success_url=reverse_lazy('password_reset_complete')
    ), name='password_reset_confirm'),

    path('reset/done', auth_views.PasswordResetCompleteView.as_view(
        template_name='dj_accounts/password_reset_complete.html'
    ), name='password_reset_complete'),
]



from django.test import TestCase, override_settings, Client

from .factories import UserFactory


class MultipleAuthenticationBackendTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory()

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_logs_user_in_with_user_email(self):
        self.client.login(email=self.user.email, password="secret")
        self.assertIn("sessionid", self.client.cookies)

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_logs_user_in_with_username_field_form_kwargs(self):
        self.client.login(email=self.user.email, password="secret")
        self.assertIn("sessionid", self.client.cookies)

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_logs_user_in_with_username(self):
        self.client.login(username=self.user.username, password="secret")
        self.assertIn("sessionid", self.client.cookies)

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_logs_user_in_with_phone(self):
        self.client.login(phone=self.user.phone, password="secret")
        self.assertIn("sessionid", self.client.cookies)

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_fails_if_username_email_or_phone_are_not_provided(self):
        self.client.login(password="secret")
        self.assertNotIn("sessionid", self.client.cookies)

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_fails_if_password_is_not_provided(self):
        self.client.login(email=self.user.email)
        self.assertNotIn("sessionid", self.client.cookies)

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_fails_if_password_is_not_correct(self):
        self.client.login(email=self.user.email, password="lol")
        self.assertNotIn("sessionid", self.client.cookies)

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_fails_if_user_is_not_active(self):
        self.user.is_active = False
        self.user.save()
        self.client.login(email=self.user.email, password="secret")
        self.assertNotIn("sessionid", self.client.cookies)

    @override_settings(AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_fails_if_user_does_not_exists(self):
        self.client.login(email="doesnotexist@mail.com", password="secret")
        self.assertNotIn("sessionid", self.client.cookies)



import inspect

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.views import LoginView
from django.contrib.auth.views import LoginView as BaseLoginView
from django.test import TestCase, Client
from django.test import override_settings
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views import View

from .factories import UserFactory
from ..forms import MultipleLoginForm, RegisterForm, UserCreationForm, UpdateUserDataForm
from ..utils import account_activation_token
from ..views import LoginView, RegisterView, VerifyPhoneView, PhoneVerificationCompleteView, VerifyEmailView, \
    EmailVerificationCompleteView, UpdateProfileView, ChangePhoneView, ResendEmailConfirmationLinkView, \
    ResendPhoneConfirmationView

UserModel = get_user_model()


class TestLoginForm(AuthenticationForm):
    pass


class LoginViewStructureTestCase(TestCase):
    def test_it_extends_django_LoginView(self):
        self.assertTrue(issubclass(LoginView, BaseLoginView))

    def test_it_has_template_name(self):
        self.assertIn('template_name', dict(inspect.getmembers(LoginView)))

    def test_template_name_is_authentication_login(self):
        self.assertEquals(LoginView.template_name, 'dj_accounts/login.html')

    def test_it_has_redirect_authenticated_user(self):
        self.assertIn('redirect_authenticated_user', dict(inspect.getmembers(LoginView)))

    def test_redirect_authenticated_user_is_true(self):
        self.assertTrue(LoginView.redirect_authenticated_user)

    def test_it_has_form_class(self):
        self.assertIn('get_form_class', dict(inspect.getmembers(LoginView)))

    def test_get_form_class_is_callable(self):
        self.assertTrue(callable(LoginView.get_form_class))


class LoginViewGetFormClassTestCase(TestCase):
    @override_settings(MULTIPLE_AUTHENTICATION_ACTIVE=False)
    def test_it_returns_default_login_form_if_MULTIPLE_AUTHENTICATION_ACTIVE_is_false_and_authentication_form_is_none(
            self):
        self.assertTrue(issubclass(LoginView().get_form_class(), AuthenticationForm))

    @override_settings(MULTIPLE_AUTHENTICATION_ACTIVE=True)
    def test_it_returns_phone_login_form_if_MULTIPLE_AUTHENTICATION_ACTIVE_is_true(self):
        self.assertTrue(issubclass(LoginView().get_form_class(), MultipleLoginForm))

    @override_settings(LOGIN_FORM=TestLoginForm)
    def test_it_returns_settings_login_form_if_is_set(self):
        self.assertTrue(issubclass(LoginView().get_form_class(), TestLoginForm))


class LoginViewPOSTTestCase(TestCase):
    def test_it_redirects_to_settings_login_redirect_url_if_user_is_logged_in(self):
        client = Client()
        user = UserFactory()
        client.login(email=user.email, password="secret")
        response = client.post(reverse('login'))
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)


class RegisterViewStructureTestCase(TestCase):
    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(RegisterView, View))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(RegisterView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(RegisterView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(RegisterView.get)[0]
        self.assertEquals(actual_signature, expected_signature)

    def test_view_has_method_post(self):
        self.assertTrue(hasattr(RegisterView, 'post'))

    def test_view_has_method_post_is_callable(self):
        self.assertTrue(callable(RegisterView.post))

    def test_post_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(RegisterView.post)[0]
        self.assertEquals(actual_signature, expected_signature)

    def test_view_has_method_get_form_class(self):
        self.assertTrue(hasattr(RegisterView, 'get_form_class'))

    def test_view_has_method_get_form_class_is_callable(self):
        self.assertTrue(callable(RegisterView.get_form_class))

    def test_get_form_class_method_signature(self):
        expected_signature = ['self']
        actual_signature = inspect.getfullargspec(RegisterView.get_form_class)[0]
        self.assertEquals(actual_signature, expected_signature)


class RegisterViewGetFormClassTestCase(TestCase):
    def test_it_returns_django_user_creation_form_if_settings_register_from_is_not_set(
            self):
        self.assertTrue(issubclass(RegisterView().get_form_class(), UserCreationForm))

    @override_settings(REGISTER_FORM=RegisterForm)
    def test_it_returns_django_user_creation_form_if_settings_register_from_is_set(self):
        self.assertTrue(issubclass(RegisterView().get_form_class(), RegisterForm))


class RegisterViewGETTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.url = reverse('register')

    def test_it_redirects_to_settings_login_redirect_url_if_user_is_logged_in(self):
        user = UserFactory()
        self.client.login(email=user.email, password="secret")
        response = self.client.get(self.url)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)

    def test_it_returns_register_template(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response, "dj_accounts/register.html")

    def test_it_returns_form_in_response_context(self):
        response = self.client.get(self.url)
        self.assertIn('form', response.context)

    @override_settings(REGISTER_FORM=RegisterForm)
    def test_response_context_form_is_instance_of_register_form_if_settings_register_from_is_set(self):
        response = self.client.get(self.url)
        self.assertIsInstance(response.context['form'], RegisterForm)


class RegisterViewPOSTTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.url = reverse('register')
        self.data = {
            "email": "test@test.test",
            "username": "TestUser",
            # "phone": "+201005263988",
            "password1": "newTESTPasswordD",
            "password2": "newTESTPasswordD",
        }

    def test_it_redirects_to_settings_login_redirect_url_if_user_is_logged_in(self):
        user = UserFactory()
        self.client.login(email=user.email, password="secret")
        response = self.client.post(self.url, self.data)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)

    def test_it_creates_user_with_provided_data(self):
        self.client.post(self.url, self.data)
        self.assertTrue(get_user_model().objects.filter(username="TestUser").exists())

    def test_it_logs_in_the_created_user(self):
        self.client.post(self.url, self.data)
        self.assertIn('_auth_user_id', self.client.session)
        self.assertEquals(self.client.session['_auth_user_id'],
                          str(get_user_model().objects.get(username="TestUser").id))

    @override_settings(LOGIN_REDIRECT_URL="/dj_accounts/login/")
    def test_it_redirect_to_next_if_next_in_request(self):
        self.data.update({"next": '/'})
        response = self.client.post(self.url, self.data)
        self.assertRedirects(response, '/', fetch_redirect_response=False)

    def test_it_redirects_to_settings_login_redirect_url_if_next_is_not_provided(self):
        response = self.client.post(self.url, self.data)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)

    def test_it_returns_register_template_if_form_is_invalid(self):
        response = self.client.post(self.url)
        self.assertTemplateUsed(response, "dj_accounts/register.html")

    def test_it_returns_form_in_response_context_if_form_is_invalid(self):
        response = self.client.post(self.url)
        self.assertIn('form', response.context)

    @override_settings(REGISTER_FORM=RegisterForm)
    def test_response_context_form_is_instance_of_register_form_if_settings_register_from_is_set(self):
        response = self.client.get(self.url)
        self.assertIsInstance(response.context['form'], RegisterForm)


# phone views
class VerifyPhoneViewStructureTestCase(TestCase):
    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(VerifyPhoneView, View))

    def test_it_extends_login_required_mixin(self):
        self.assertTrue(issubclass(VerifyPhoneView, LoginRequiredMixin))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(VerifyPhoneView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(VerifyPhoneView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(VerifyPhoneView.get)[0]
        self.assertEquals(actual_signature, expected_signature)

    def test_view_has_method_post(self):
        self.assertTrue(hasattr(VerifyPhoneView, 'post'))

    def test_view_has_method_post_is_callable(self):
        self.assertTrue(callable(VerifyPhoneView.post))

    def test_post_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(VerifyPhoneView.post)[0]
        self.assertEquals(actual_signature, expected_signature)


class VerifyPhoneViewGETTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory()
        self.client.login(email=self.user.email, password="secret")
        self.url = reverse('verify-phone')

    def test_it_redirects_to_login_if_user_is_not_logged_in(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse("login") + "?next=" + self.url, fetch_redirect_response=False)

    def test_it_redirects_to_settings_login_redirect_url_if_phone_is_verified(self):
        user = UserFactory(phone_verified_at=now())
        self.client.login(email=user.email, password="secret")
        response = self.client.get(self.url)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)

    def test_it_returns_verify_phone_template(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response, "dj_accounts/verify_phone.html")

    def test_it_returns_form_in_response_context(self):
        response = self.client.get(self.url)
        self.assertIn('form', response.context)


class VerifyPhoneViewPOSTTestCase(TestCase):
    def setUp(self):
        self.user = UserFactory(phone="201002536987")
        self.client.login(email=self.user.email, password="secret")
        self.url = reverse('verify-phone')
        self.data = {"code": "777777"}

    def test_it_redirects_to_login_if_user_is_not_logged_in(self):
        self.client.logout()
        response = self.client.post(self.url)
        self.assertRedirects(response, reverse("login") + "?next=" + self.url, fetch_redirect_response=False)

    def test_it_redirects_to_settings_login_redirect_url_if_phone_is_verified(self):
        user = UserFactory(phone_verified_at=now())
        self.client.login(email=user.email, password="secret")
        response = self.client.post(self.url)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)

    def test_it_updates_phone_verified_at_column_in_user_model_to_now_on_success(self):
        self.client.post(self.url, self.data)
        self.user.refresh_from_db()
        self.assertIsNotNone(self.user.phone_verified_at)

    def test_it_redirects_to_phone_verification_complete_on_success(self):
        response = self.client.post(self.url, self.data)
        self.assertRedirects(response, reverse("phone-verification-complete"), fetch_redirect_response=False)

    def test_it_returns_verify_phone_template_on_failure(self):
        response = self.client.post(self.url)
        self.assertTemplateUsed(response, "dj_accounts/verify_phone.html")

    def test_it_returns_form_in_response_context_on_failure(self):
        response = self.client.post(self.url)
        self.assertIn('form', response.context)


class PhoneVerificationCompleteViewStructureTestCase(TestCase):
    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(PhoneVerificationCompleteView, View))

    def test_it_extends_login_required_mixin(self):
        self.assertTrue(issubclass(PhoneVerificationCompleteView, LoginRequiredMixin))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(PhoneVerificationCompleteView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(PhoneVerificationCompleteView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(PhoneVerificationCompleteView.get)[0]
        self.assertEquals(actual_signature, expected_signature)


class PhoneVerificationCompleteViewGETTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory()
        self.client.login(email=self.user.email, password="secret")
        self.url = reverse('phone-verification-complete')

    def test_it_redirects_to_login_if_user_is_not_logged_in(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse("login") + "?next=" + self.url)

    def test_it_returns_phone_verification_complete_template(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response, "dj_accounts/phone_verification_complete.html")


class ResendPhoneConfirmationViewStructureTestCase(TestCase):
    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(ResendPhoneConfirmationView, View))

    def test_it_extends_login_required_mixin(self):
        self.assertTrue(issubclass(ResendPhoneConfirmationView, LoginRequiredMixin))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(ResendPhoneConfirmationView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(ResendPhoneConfirmationView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(ResendPhoneConfirmationView.get)[0]
        self.assertEquals(actual_signature, expected_signature)


class ResendPhoneConfirmationViewTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory()
        self.client.login(email=self.user.email, password="secret")
        self.url = reverse("resend_phone_activation")

    def test_it_redirects_to_login_if_user_is_not_logged_in(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse("login") + "?next=" + self.url)

    def test_it_redirect_to_phone_verification_again(self):
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse("verify-phone"))

    def test_message_is_correct(self):
        response = self.client.get(self.url, follow=True)
        msgs = list(messages.get_messages(response.wsgi_request))
        self.assertEquals(str(msgs[0]), _("A new confirmation code has been sent to your phone"))


# Email Verification


class VerifyEmailViewStructureTestCase(TestCase):

    def test_it_extends_login_required_mixin(self):
        self.assertTrue(issubclass(VerifyEmailView, LoginRequiredMixin))

    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(VerifyEmailView, View))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(VerifyEmailView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(VerifyEmailView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request', 'uidb64', 'token']
        actual_signature = inspect.getfullargspec(VerifyEmailView.get)[0]
        self.assertEquals(actual_signature, expected_signature)


class VerifyEmailViewTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory()
        self.client.login(email=self.user.email, password="secret")
        self.user = UserFactory()
        self.uid = urlsafe_base64_encode(force_bytes(self.user.pk))
        self.token = account_activation_token.make_token(self.user)

    def test_it_activate_the_user(self):
        confirm = self.client.get(reverse('verify-email', args=[self.uid, self.token]))
        self.user.refresh_from_db()
        self.assertIsNotNone(self.user.email_verified_at)

    def test_it_redirects_to_phone_verification_complete_on_success(self):
        confirm = self.client.get(reverse('verify-email', args=[self.uid, self.token]))
        self.assertRedirects(confirm, reverse('email-verification-complete'))


class EmailVerificationCompleteViewStructureTestCase(TestCase):
    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(EmailVerificationCompleteView, View))

    def test_it_extends_login_required_mixin(self):
        self.assertTrue(issubclass(EmailVerificationCompleteView, LoginRequiredMixin))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(EmailVerificationCompleteView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(EmailVerificationCompleteView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(EmailVerificationCompleteView.get)[0]
        self.assertEquals(actual_signature, expected_signature)


class EmailVerificationCompleteViewGETTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory()
        self.client.login(email=self.user.email, password="secret")
        self.url = reverse('email-verification-complete')

    def test_it_redirects_to_login_if_user_is_not_logged_in(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse("login") + "?next=" + self.url, fetch_redirect_response=False)

    def test_it_returns_phone_verification_complete_template(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response, "dj_accounts/email_verification_complete.html")


class ResendEmailConfirmationLinkViewStructureTestCase(TestCase):
    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(ResendEmailConfirmationLinkView, View))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(ResendEmailConfirmationLinkView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(ResendEmailConfirmationLinkView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(ResendEmailConfirmationLinkView.get)[0]
        self.assertEquals(actual_signature, expected_signature)


# profile views

class UpdateProfileInfoViewStructureTestCase(TestCase):

    @override_settings(PROFILE_FORM=RegisterForm)
    def setUp(self):
        pass

    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(UpdateProfileView, View))

    def test_it_extends_login_required_mixin(self):
        self.assertTrue(issubclass(UpdateProfileView, LoginRequiredMixin))

    def test_view_has_method_get_form_class(self):
        self.assertTrue(hasattr(UpdateProfileView, 'get_form_class'))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(UpdateProfileView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(UpdateProfileView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(UpdateProfileView.get)[0]
        self.assertEquals(actual_signature, expected_signature)

    def test_view_has_method_post(self):
        self.assertTrue(hasattr(UpdateProfileView, 'post'))

    def test_view_has_method_post_is_callable(self):
        self.assertTrue(callable(UpdateProfileView.post))

    def test_post_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(UpdateProfileView.post)[0]
        self.assertEquals(actual_signature, expected_signature)


class UpdateProfileInfoGetViewTestCase(TestCase):

    @override_settings(PROFILE_FORM=UpdateUserDataForm)
    def setUp(self):
        self.client = Client()
        self.user = UserFactory()
        self.client.login(username=self.user.username, password='secret')
        self.url = reverse('update-profile')

    def test_it_redirects_to_login_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse('login') + '?next=' + self.url)

    def test_template_used(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response, "dj_accounts/update_user_data_form.html")

    def test_context_contains_form(self):
        response = self.client.get(self.url)
        self.assertIn("form", response.context)

    def test_context_contains_page_title(self):
        response = self.client.get(self.url)
        self.assertIn("page_title", response.context)

    def test_context_page_title_is_instance_of_Update_form(self):
        response = self.client.get(self.url)
        self.assertEquals(response.context['page_title'], _("Update User Info"))


class UpdateProfileInfoPostViewTestCase(TestCase):

    def setUp(self):
        self.client = Client()
        self.user = UserFactory(first_name='abc', last_name='xyz')
        self.client.login(email=self.user.email, password='secret')
        self.url = reverse('update-profile')
        self.data = {
            "first_name": "Test",
            "last_name": "User",
            'email': 'test@test.com',
        }

    def test_it_redirects_to_login_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.post(self.url, data=self.data)
        self.assertRedirects(response, reverse('login') + '?next=' + self.url)

    def test_template_used(self):
        response = self.client.post(self.url)
        self.assertTemplateUsed(response, "dj_accounts/update_user_data_form.html")

    def test_context_contains_form_on_error(self):
        response = self.client.post(self.url, data={})
        self.assertIn("form", response.context)

    @override_settings(PROFILE_FORM=UpdateUserDataForm)
    def test_it_redirect_to_success_url(self):
        response = self.client.post(self.url, data=self.data)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)

    @override_settings(PROFILE_FORM=UpdateUserDataForm)
    def test_message_value_when_data_posted_successfully(self):
        response = self.client.post(self.url, data=self.data)
        msgs = list(messages.get_messages(response.wsgi_request))
        self.assertEquals(str(msgs[0]), _('your profile has been updated successfully'))


class ChangeEmailViewGetTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory(username='test', password='secret')
        self.client.login(username=self.user.username, password='secret')
        self.url = reverse('change-email')

    def test_it_redirects_to_login_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse('login') + '?next=' + self.url, fetch_redirect_response=False)

    def test_template_used(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response, "dj_accounts/update_email_form.html")

    def test_context_contains_form(self):
        response = self.client.get(self.url)
        self.assertIn("form", response.context)

    def test_context_contains_page_title(self):
        response = self.client.get(self.url)
        print(response.context)

        self.assertIn("page_title", response.context)

    def test_context_page_title_is_instance_of_Update_form(self):
        response = self.client.get(self.url)
        self.assertEquals(response.context['page_title'], _("Update Email"))


class ChangeEmailViewPostTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory(username='test@test.test', password='secret')
        self.client.login(username=self.user.username, password='secret')
        self.url = reverse('change-email')
        self.data = {
            "new_email": 'test101@test.test',
            "password": "secret",
        }

    def test_it_redirects_to_login_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.post(self.url, data=self.data)
        self.assertRedirects(response, reverse('login') + '?next=' + self.url)

    def test_template_used(self):
        response = self.client.post(self.url)
        self.assertTemplateUsed(response, "dj_accounts/update_email_form.html")

    def test_context_contains_form(self):
        response = self.client.post(self.url)
        self.assertIn("form", response.context)

    def test_it_redirect_to_success_url(self):
        response = self.client.post(self.url, data=self.data)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)


class ChangePhoneViewGETTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory(phone="+201005263988")
        self.client.login(username=self.user.username, password='secret')
        self.url = reverse('change-phone')

    def test_it_redirects_to_login_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse('login') + '?next=' + self.url)

    def test_template_used(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response, "dj_accounts/update_phone_number_form.html")

    def test_context_contains_form(self):
        response = self.client.get(self.url)
        self.assertIn("form", response.context)

    def test_context_contains_page_title(self):
        response = self.client.get(self.url)
        self.assertIn("page_title", response.context)

    def test_context_page_title_is_instance_of_Update_form(self):
        response = self.client.get(self.url)
        self.assertEquals(response.context['page_title'], _("Update Phone Number"))


class ChangePhoneViewTestCase(TestCase):
    def test_it_extends_django_class_base_view(self):
        self.assertTrue(issubclass(ChangePhoneView, View))

    def test_it_extends_django_login_required_mixin(self):
        self.assertTrue(issubclass(ChangePhoneView, LoginRequiredMixin))

    def test_it_has_get_attribute(self):
        self.assertTrue(hasattr(ChangePhoneView, 'get'))

    def test_it_has_post_attribute(self):
        self.assertTrue(hasattr(ChangePhoneView, 'post'))

    def test_get_attribute_is_callable(self):
        self.assertTrue(callable(ChangePhoneView.get))

    def test_post_attribute_is_callable(self):
        self.assertTrue(callable(ChangePhoneView.post))

    def test_get_signature(self):
        expected_signature = ['self', 'request', ]
        actual_signature = inspect.getfullargspec(ChangePhoneView.get)[0]
        self.assertEquals(actual_signature, expected_signature)

    def test_post_signature(self):
        expected_signature = ['self', 'request', ]
        actual_signature = inspect.getfullargspec(ChangePhoneView.post)[0]
        self.assertEquals(actual_signature, expected_signature)


class ChangePhoneViewPOSTTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = UserFactory(phone="+201001452369")
        self.client.login(username=self.user.username, password='secret')
        self.url = reverse('change-phone')
        self.data = {
            "new_phone": '+201005263977',
            "password": "secret",
        }

    def test_it_redirects_to_login_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.post(self.url, data=self.data)
        self.assertRedirects(response, reverse('login') + '?next=' + self.url)

    def test_template_used(self):
        response = self.client.post(self.url)
        self.assertTemplateUsed(response, "dj_accounts/update_phone_number_form.html")

    def test_context_contains_form(self):
        response = self.client.post(self.url)
        self.assertIn("form", response.context)

    def test_it_rerender_form_on_failure(self):
        UserFactory(phone='+201005263977')
        response = self.client.post(self.url, self.data)
        self.assertTemplateUsed(response, "dj_accounts/update_phone_number_form.html")

    def test_it_redirect_to_success_url(self):
        response = self.client.post(self.url, data=self.data)
        self.assertRedirects(response, settings.LOGIN_REDIRECT_URL, fetch_redirect_response=False)



from django.test import TestCase

from .factories import UserFactory
from ..utils import account_activation_token


class TestAccountActivationToken(TestCase):

    def setUp(self):
        self.user = UserFactory()

    def test_it_create_token_for_the_user(self):
        token = account_activation_token.make_token(self.user)
        self.assertGreater(len(token), 0)

    def test_it_valid_token(self):
        token = account_activation_token.make_token(self.user)
        self.assertTrue(account_activation_token.check_token(self.user, token))



import random

import factory
from django.contrib.auth import get_user_model
from factory.django import DjangoModelFactory

UserModel = get_user_model()


class UserFactory(DjangoModelFactory):
    class Meta:
        model = UserModel
        django_get_or_create = ('phone',)

    first_name = factory.Faker('first_name')
    last_name = factory.Faker('last_name')
    username = factory.lazy_attribute(
        lambda obj: '{}.{}.{}'.format(obj.first_name, obj.last_name, random.randrange(1, 1000)))
    email = factory.lazy_attribute(lambda obj: '{}@email.com'.format(obj.username))
    phone = factory.Sequence(lambda n: '3215616_%d' % n)
    password = factory.PostGenerationMethodCall('set_password', 'secret')



import importlib
import sys
import traceback

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import PasswordResetForm, PasswordChangeForm
from django.contrib.auth.tokens import default_token_generator
from django.utils.encoding import force_str
from django.utils.http import urlsafe_base64_decode
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy as _
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_simplejwt.exceptions import TokenError
from rest_framework_simplejwt.tokens import RefreshToken

from .forms import UpdateEmailForm, UpdatePhoneNumberForm, MultipleLoginForm
from .forms import VerifyPhoneForm
from .utils import account_activation_token, send_mail_confirmation, get_user_tokens, get_errors, get_settings_value, \
    get_class_from_settings
from .verify_phone import VerifyPhone

UserModel = get_user_model()


class RegisterAPIView(APIView):
    access_token = None
    refresh_token = None
    token = None
    user = None
    authentication_classes = []
    permission_classes = []

    def get_form_class(self):
        form_class = getattr(settings, "REGISTER_FORM", 'dj_accounts.forms.UserCreationForm')
        if type(form_class) is str:
            form_class_split = form_class.split('.')
            class_name = form_class_split[-1:][0]
            module_name = form_class_split[:-1]
            return getattr(importlib.import_module('.'.join(module_name)), class_name)
        else:
            return form_class

    def post(self, request, *args, **kwargs):
        form = self.get_form_class()(data=request.data)
        if form.is_valid():
            self.user = form.save()

            try:
                send_mail_confirmation(request, self.user)
            except Exception as e:
                parts = ["Traceback (most recent call last):\n"]
                parts.extend(traceback.format_stack(limit=25)[:-2])
                parts.extend(traceback.format_exception(*sys.exc_info())[1:])
                print("".join(parts))

            try:
                VerifyPhone().send(self.user.phone)
            except Exception as e:
                parts = ["Traceback (most recent call last):\n"]
                parts.extend(traceback.format_stack(limit=25)[:-2])
                parts.extend(traceback.format_exception(*sys.exc_info())[1:])
                print("".join(parts))

            refresh = RefreshToken.for_user(self.user)
            return Response({
                "access_token": str(refresh.access_token),
                "refresh_token": str(refresh)
            }, status=status.HTTP_201_CREATED)
        errors = {name: error[0] for name, error in form.errors.as_data().items()}
        return Response(errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class LoginAPIView(APIView):
    authentication_classes = []
    permission_classes = []

    def get_form_class(self):
        """
        if MULTIPLE_AUTHENTICATION_ACTIVE is True
        - returns MultipleLoginForm
        else
        - returns the regular AuthenticationForm if LOGIN_FORM setting is not defined
        - returns the provided LOGIN_FORM if set
        """
        if get_settings_value('MULTIPLE_AUTHENTICATION_ACTIVE', False):
            return MultipleLoginForm
        return get_class_from_settings('LOGIN_FORM', 'django.contrib.auth.forms.AuthenticationForm')

    def post(self, request, *args, **kwargs):
        form = self.get_form_class()(data=request.data)
        if form.is_valid():
            user = form.user_cache
            tokens = get_user_tokens(user)
            return Response(tokens, status=status.HTTP_200_OK)

        return Response(get_errors(form.errors.as_data()), status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class UserLogoutAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        token = request.POST.get('refresh', None)
        if not token:
            return Response({_("token field is required")}, status=status.HTTP_422_UNPROCESSABLE_ENTITY)

        try:
            RefreshToken(token).blacklist()
        except TokenError:
            raise ValidationError({"refresh": _('Invalid token.')})

        return Response(status=status.HTTP_204_NO_CONTENT)


class PasswordResetAPIView(APIView):
    permission_classes = []
    authentication_classes = []
    email_template_name = 'dj_accounts/password_reset_email.html'
    extra_email_context = None
    from_email = None
    html_email_template_name = None
    subject_template_name = 'dj_accounts/password_reset_subject.txt'
    token_generator = default_token_generator

    def post(self, request, *args, **kwargs):
        form = PasswordResetForm(data=request.data)
        if form.is_valid():
            opts = {
                'use_https': self.request.is_secure(),
                'token_generator': self.token_generator,
                'from_email': self.from_email,
                'email_template_name': self.email_template_name,
                'subject_template_name': self.subject_template_name,
                'request': self.request,
                'html_email_template_name': self.html_email_template_name,
                'extra_email_context': self.extra_email_context,
            }
            form.save(**opts)
            return Response({"message": "{} {}".format(
                _('We’ve emailed you instructions for setting your password, '
                  'if an account exists with the email you entered. You should receive them shortly.'),
                _('If you don’t receive an email, please make sure you’ve entered '
                  'the address you registered with, and check your spam folder.'))},
                status=status.HTTP_200_OK)
        return Response(get_errors(form.errors.as_data()), status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class VerifyPhoneAPIView(APIView):
    permission_classes = [IsAuthenticated, ]

    def post(self, request, *args, **kwargs):
        if request.user.phone_verified_at is not None:
            return Response({'message': 'this account was activated before'}, status=status.HTTP_400_BAD_REQUEST)

        form = VerifyPhoneForm(request.POST, user=request.user)
        if form.is_valid():
            request.user.phone_verified_at = now()
            request.user.save()
            return Response(status=status.HTTP_200_OK)
        return Response(form.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class ResendPhoneConfirmationAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        try:
            VerifyPhone().send(request.user.phone)
        except Exception as e:
            parts = ["Traceback (most recent call last):\n"]
            parts.extend(traceback.format_stack(limit=25)[:-2])
            parts.extend(traceback.format_exception(*sys.exc_info())[1:])
            print("".join(parts))

        return Response({"message": _('Code was resent successfully.')}, status=status.HTTP_200_OK)


# email verification
class ResendEmailConfirmationLinkAPIView(APIView):
    def get(self, request, *args, **kwargs):
        try:
            send_mail_confirmation(request, request.user)
        except Exception as e:
            parts = ["Traceback (most recent call last):\n"]
            parts.extend(traceback.format_stack(limit=25)[:-2])
            parts.extend(traceback.format_exception(*sys.exc_info())[1:])
            print("".join(parts))

        return Response({"message": _('Email activation link resent successfully')})


class VerifyEmailAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, uidb64, token):
        try:
            uid = force_str(urlsafe_base64_decode(uidb64))
            user = UserModel.objects.get(pk=uid)
        except(TypeError, ValueError, OverflowError, UserModel.DoesNotExist):
            user = None
        if user is not None and account_activation_token.check_token(user, token):
            user.email_verified_at = now()
            user.save()
        return Response({"message": _('Email was verified successfully.')}, status=status.HTTP_200_OK)


class ProfileDetailsAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def get_profile_serializer(self):
        profile_handler = getattr(settings, "PROFILE_SERIALIZER_HANDLER",
                                  'dj_accounts.tests.profile_handlers.profile_serializer_handler')
        profile_handler_split = profile_handler.split('.')
        class_name = profile_handler_split[-1:][0]
        module_name = profile_handler_split[:-1]
        return getattr(importlib.import_module('.'.join(module_name)), class_name)()

    def get(self, request, *args, **kwargs):
        return Response(self.get_profile_serializer()(self.request.user).data, status=status.HTTP_200_OK)


class UpdateProfileAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def get_form_class(self):
        form_class = getattr(settings, "PROFILE_FORM", 'dj_accounts.forms.UserChangeForm')
        if type(form_class) is str:
            form_class_split = form_class.split('.')
            class_name = form_class_split[-1:][0]
            module_name = form_class_split[:-1]
            return getattr(importlib.import_module('.'.join(module_name)), class_name)
        return form_class

    def get_profile_serializer(self):
        profile_handler = getattr(settings, "PROFILE_SERIALIZER_HANDLER",
                                  'dj_accounts.tests.profile_handlers.profile_serializer_handler')
        profile_handler_split = profile_handler.split('.')
        class_name = profile_handler_split[-1:][0]
        module_name = profile_handler_split[:-1]
        return getattr(importlib.import_module('.'.join(module_name)), class_name)()

    def post(self, request, *args, **kwargs):
        form = self.get_form_class()(data=request.data, instance=request.user)
        if form.is_valid():
            form.save()
            self.request.user.refresh_from_db()
            return Response(self.get_profile_serializer()(self.request.user).data, status=status.HTTP_200_OK)
        return Response(form.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class ChangePasswordAPIView(APIView):
    permission_classes = [IsAuthenticated]

    def post(self, request, *args, **kwargs):
        form = PasswordChangeForm(data=request.data, user=request.user)
        if form.is_valid():
            form.save()
            return Response(data={'message': _("Password updated successfully")}, status=status.HTTP_200_OK)
        return Response(get_errors(form.errors.as_data()), status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class ChangeEmailAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        form = UpdateEmailForm(data=request.data, user=request.user)
        if form.is_valid():
            form.save()
            return Response({'message': _('Email Changed Successfully')}, status=status.HTTP_200_OK)
        return Response(get_errors(form.errors.as_data()), status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class ChangePhoneAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        form = UpdatePhoneNumberForm(data=request.data, user=request.user)
        if form.is_valid():
            form.save()
            return Response({'message': _('Phone Changed Successfully')}, status=status.HTTP_200_OK)
        return Response(get_errors(form.errors.as_data()), status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class DeleteProfileAPIView(APIView):
    permission_classes = [IsAuthenticated]

    def post(self, request, *args, **kwargs):
        UserLogoutAPIView().post(request, *args, **kwargs)
        request.user.delete()
        return Response({
            "message": _("Account Deleted Successfully!")
        })



from django.urls import path

from ...views_api import ChangeEmailAPIView, UpdateProfileAPIView, ChangePhoneAPIView, ChangePasswordAPIView, \
    ProfileDetailsAPIView, DeleteProfileAPIView

urlpatterns = [
    path('', ProfileDetailsAPIView.as_view(), name="user-profile"),
    path('update/password/', ChangePasswordAPIView.as_view(), name='change_password_api'),
    path('update/profile/', UpdateProfileAPIView.as_view(), name='update-profile-api'),
    path('update/email/', ChangeEmailAPIView.as_view(), name='change-email-api'),
    path('update/phone/', ChangePhoneAPIView.as_view(), name='change-phone-api'),
    path('delete/', DeleteProfileAPIView.as_view(), name='delete-profile-api'),
]



from ...views_api import VerifyPhoneAPIView, \
    VerifyEmailAPIView, ResendPhoneConfirmationAPIView, RegisterAPIView, ResendEmailConfirmationLinkAPIView, \
    UserLogoutAPIView, PasswordResetAPIView, LoginAPIView, ChangePasswordAPIView
from django.urls import path
from rest_framework_simplejwt.views import TokenRefreshView

urlpatterns = [

    # simple jwt
    path('login/', LoginAPIView.as_view(), name='login_api'),
    path('token/refresh/', TokenRefreshView.as_view(), name='token_refresh'),
    path('logout/', UserLogoutAPIView.as_view(), name='logout_api'),
    path('signup/', RegisterAPIView.as_view(), name='signup_api'),

    # password
    path('password_reset/', PasswordResetAPIView.as_view(), name='password_reset_api'),

    # email verification
    path('verify/email/<str:uidb64>/<str:token>/', VerifyEmailAPIView.as_view(), name='verify_email_api'),
    path('resend_email_activation/', ResendEmailConfirmationLinkAPIView.as_view(), name='resend_email_activation_api'),

    # phone verification
    path('verify/phone/', VerifyPhoneAPIView.as_view(), name='verify_phone_api'),
    path('resend_phone_activation/', ResendPhoneConfirmationAPIView.as_view(), name='resend_phone_activation_api'),
]
